// C++ implementation of galois field arithmetic
// January 1999
// Copyright 1999 Phil Karn
// May be used under the terms of the GNU Public License (GPL)

#include "galois.h"

/* CCSDS field generator polynomial: 1+x+x^2+x^7+x^8 */
int Pp[MM+1] = { 1, 1, 1, 0, 0, 0, 0, 1, 1 };

/* index->polynomial form conversion table */
int Alpha_to[NN + 1];

/* Polynomial->index form conversion table */
int Index_of[NN + 1];


/* generate GF(2**m) from the irreducible polynomial p(X) in Pp[0]..Pp[m]
   lookup tables:  index->polynomial form   alpha_to[] contains j=alpha**i;
                   polynomial form -> index form  index_of[j=alpha**i] = i
   alpha=2 is the primitive element of GF(2**m)
   HARI's COMMENT: (4/13/94) alpha_to[] can be used as follows:
        Let @ represent the primitive element commonly called "alpha" that
   is the root of the primitive polynomial p(x). Then in GF(2^m), for any
   0 <= i <= 2^m-2,
        @^i = a(0) + a(1) @ + a(2) @^2 + ... + a(m-1) @^(m-1)
   where the binary vector (a(0),a(1),a(2),...,a(m-1)) is the representation
   of the integer "alpha_to[i]" with a(0) being the LSB and a(m-1) the MSB. Thus for
   example the polynomial representation of @^5 would be given by the binary
   representation of the integer "alpha_to[5]".
                   Similarily, index_of[] can be used as follows:
        As above, let @ represent the primitive element of GF(2^m) that is
   the root of the primitive polynomial p(x). In order to find the power
   of @ (alpha) that has the polynomial representation
        a(0) + a(1) @ + a(2) @^2 + ... + a(m-1) @^(m-1)
   we consider the integer "i" whose binary representation with a(0) being LSB
   and a(m-1) MSB is (a(0),a(1),...,a(m-1)) and locate the entry
   "index_of[i]". Now, @^index_of[i] is that element whose polynomial 
    representation is (a(0),a(1),a(2),...,a(m-1)).
   NOTE:
        The element alpha_to[2^m-1] = 0 always signifying that the
   representation of "@^infinity" = 0 is (0,0,0,...,0).
        Similarily, the element index_of[0] = A0 always signifying
   that the power of alpha which has the polynomial representation
   (0,0,...,0) is "infinity".
 
*/

static void
generate_gf(void)
{
  register int i, mask;

  mask = 1;
  Alpha_to[MM] = 0;
  for (i = 0; i < MM; i++) {
    Alpha_to[i] = mask;
    Index_of[Alpha_to[i]] = i;
    /* If Pp[i] == 1 then, term @^i occurs in poly-repr of @^MM */
    if (Pp[i] != 0)
      Alpha_to[MM] ^= mask;	/* Bit-wise EXOR operation */
    mask <<= 1;	/* single left-shift */
  }
  Index_of[Alpha_to[MM]] = MM;
  /*
   * Have obtained poly-repr of @^MM. Poly-repr of @^(i+1) is given by
   * poly-repr of @^i shifted left one-bit and accounting for any @^MM
   * term that may occur when poly-repr of @^i is shifted.
   */
  mask >>= 1;
  for (i = MM + 1; i < NN; i++) {
    if (Alpha_to[i - 1] >= mask)
      Alpha_to[i] = Alpha_to[MM] ^ ((Alpha_to[i - 1] ^ mask) << 1);
    else
      Alpha_to[i] = Alpha_to[i - 1] << 1;
    Index_of[Alpha_to[i]] = i;
  }
  Index_of[0] = A0;
  Alpha_to[NN] = 0;
}

// zero in index form
const int A0(NN);

// Primitive root alpha, can be used in expressions like
// ALPHA^5 = alpha to the fifth power
//const galois ALPHA(galois_index(1));
const galois ALPHA(2); // ****EXPEDIENT HACK, FIX THIS*****

// Constant X that may be used to construct polynomials by hand, e.g.,
// (X<<2) + 2*(X<<1) + 3*(X) + 1
const polynomial X(1,POLY,0,1);

void
init_galois()
{
  generate_gf();
}

// display polynomial in polynomial form
void dump_poly(const polynomial& p,const char *label) {
  int i;
  
  printf("%s(%d), poly form:",label,p.degree());
  for(i=0;i<=p.degree();i++)
    printf(" %d",p[i].poly());
  printf("\n");
}
// display in index form
void dump_index(const polynomial& p,const char *label) {
  int i;
  
  printf("%s(%d), index form:",label,p.degree());
  for(i=0;i<=p.degree();i++)
    printf(" %d",p[i].index());
  printf("\n");
}

// Comparision
int operator ==(const polynomial& arg1,const polynomial& arg2) {
  int i;
  if(arg1.degree() != arg2.degree())
    return 0;
  for(i=arg1.degree();i>=0;i--)
    if(arg1[i] != arg2[i])
      return 0;
  return 1;
}

// Polynomial addition
polynomial operator +(const polynomial& arg1,const polynomial& arg2) {
  int i,j;
  polynomial result(max(arg1.degree(),arg2.degree()));
  
  j = min(arg1.degree(),arg2.degree());
  for(i=0;i<=j;i++)
    result[i] = arg1[i] + arg2[i];
  if(arg2.degree() >= arg1.degree()){
    for(;i<=arg2.degree();i++){
      result[i] = arg2[i];
    }
  } else {
    for(;i<=arg1.degree();i++){
      result[i] = arg1[i];
    }
  }
  return result;
}
polynomial operator +(const polynomial& arg1,const galois& arg2) {
  int i;
  polynomial result(arg1.degree());
  result[0] = arg1[0] + arg2;
  for(i=arg1.degree();i>0;i--)
    result[i] = arg1[i];
  return result;
}
// Polynomial multiplication
// this needs to be reduced modulo X^deg+1
polynomial operator *(const polynomial& arg1,const polynomial& arg2) {
  int i,j;
  polynomial result;

  result.clear(arg1.degree() + arg2.degree());
  for(i=arg1.degree();i>=0;i--){
    for(j=arg2.degree();j>=0;j--){
      result[i+j] += arg1[i] * arg2[j];
    }
  }
  return result;
}
polynomial operator *(const polynomial& arg1,const galois& arg2) {
  int i;

  polynomial result(arg1.degree());
  for(i=arg1.degree();i>=0;i--){
    result[i] = arg1[i] * arg2;
  }
  return result;
}
// Polynomial division, returning quotient
polynomial operator /(const polynomial& dividend,const polynomial& divisor) {
  int i,j;
  polynomial rem(divisor.degree()-1);;
  polynomial quot(dividend.degree());
  for(i=dividend.degree();i>=0;i--){
    quot[i] = rem[rem.degree()] * -divisor[divisor.degree()].inv();
    for(j=rem.degree();j> 0;j--)
      rem[j] = rem[j-1] + (quot[i] * divisor[j]);
    rem[0] = dividend[i] + (quot[i] * divisor[0]);
  }
  return quot;
}
polynomial operator /(const polynomial& dividend,const galois& divisor){
  int i;
  polynomial quot(dividend.degree());
  for(i=dividend.degree();i>=0;i--){
    quot[i] = dividend[i] / divisor;
  }
  return quot;
}
// Multiply polynomial by X^n
// this needs to be reduced modulo X^deg+1
polynomial operator <<(const polynomial& arg1,const int arg2) {
  polynomial product(arg1.degree()+arg2);
  int i;
  
  for(i=arg1.degree();i >= 0;i--)
    product[i+arg2] = arg1[i];

  for(i=arg2-1;i >=0;i--)
    product[i] = 0;

  return product;
}
// divide polynomial by X^n
polynomial operator >>(const polynomial& arg1,const int arg2) {
  polynomial product(arg1.degree()-arg2);
  int i;
  
  for(i=arg1.degree()-arg2;i>=0;i--)
    product[i] = arg1[i+arg2];

  return product;
}
// Polynomial division, returning remainder
polynomial operator %(const polynomial& dividend,const polynomial& divisor) {
  int i,j;
  polynomial rem;
  galois quot; // current quotient term
  
  rem.clear(divisor.degree()-1);
  for(i=dividend.degree();i>=0;i--){
    quot = rem[rem.degree()] * -divisor[divisor.degree()].inv();
    for(j=rem.degree();j> 0;j--)
      rem[j] = rem[j-1] + (quot * divisor[j]);
    rem[0] = dividend[i] + (quot * divisor[0]);
  }
  return rem;
}
// Exponentiate polynomial - simpleminded algorithm
// this needs to be made more efficient and reduced modulo X^deg+1
polynomial operator ^(const polynomial& a, const int n) {
  int i;
  polynomial result(0);
  result[0] = 1;

  for(i=0;i<n;i++)
    result *= a;
  return result;
}
